@@ -134,7 +134,7 @@ gnome_file_entry_get_type (void)
 	static GType object_type = 0;
 
 	if (object_type == 0) {
-		static const GTypeInfo object_info = {
+		const GTypeInfo object_info = {
 			sizeof (GnomeFileEntryClass),
 			(GBaseInitFunc) NULL,
 			(GBaseFinalizeFunc) NULL,
@@ -147,7 +147,7 @@ gnome_file_entry_get_type (void)
 			NULL			/* value_table */
 		};
 
-		static const GInterfaceInfo editable_info = {
+		const GInterfaceInfo editable_info = {
 			(GInterfaceInitFunc) gnome_file_entry_editable_init,	 /* interface_init */
 			NULL,			                         	 /* interface_finalize */
 			NULL			                         	 /* interface_data */
