@@ -39,24 +39,27 @@
 #include "gnome-uidefs.h"
 
 #include "gnome-icon-sel.h"
+#include "libgnomeui-access.h"
 
-#include <libgnomevfs/gnome-vfs-mime.h>
+#include <libgnomevfs/gnome-vfs-ops.h>
 
 #define ICON_SIZE 48
 
 struct _GnomeIconSelectionPrivate {
-  GtkWidget * box;
+	GtkWidget * box;
 
-  GtkWidget * gil;
+	GtkWidget * gil;
 
-  GList * file_list;
-  
-  gboolean stop_loading; /* a flag set to stop the loading of images in midprocess */
+	GList * file_list;
+
+	/* For loading the icons */
+	GMainLoop * load_loop;
+	guint load_idle;
+	int load_i;
+	int load_file_count;
+	GtkWidget * load_progressbar;
 };
 
-static void gnome_icon_selection_class_init (GnomeIconSelectionClass *klass);
-static void gnome_icon_selection_instance_init (GnomeIconSelection   *gis);
-
 static void gnome_icon_selection_destroy    (GtkObject               *object);
 static void gnome_icon_selection_finalize   (GObject                 *object);
 
@@ -64,25 +67,24 @@ static int sort_file_list		    (gconstpointer            a,
 					     gconstpointer            b);
 
 GNOME_CLASS_BOILERPLATE (GnomeIconSelection, gnome_icon_selection,
-			 GtkVBox, gtk_vbox, GTK_TYPE_VBOX)
+			 GtkVBox, GTK_TYPE_VBOX)
 
 static void
 gnome_icon_selection_class_init (GnomeIconSelectionClass *klass)
 {
-  GtkObjectClass *object_class;
   GObjectClass *gobject_class;
+  GtkObjectClass *object_class;
 
-  object_class = (GtkObjectClass*) klass;
   gobject_class = (GObjectClass*) klass;
+  object_class = (GtkObjectClass*) klass;
 
-  object_class->destroy = gnome_icon_selection_destroy;
   gobject_class->finalize = gnome_icon_selection_finalize;
+  object_class->destroy = gnome_icon_selection_destroy;
 }
 
 static void
 gnome_icon_selection_instance_init (GnomeIconSelection *gis)
 {
-	GtkAdjustment *vadj;
 	GtkWidget *box;
 	GtkWidget *frame;
 	GtkWidget *sb;
@@ -105,19 +107,24 @@ gnome_icon_selection_instance_init (GnomeIconSelection *gis)
 	gtk_box_pack_start (GTK_BOX (box), frame, TRUE, TRUE, 0);
 	gtk_widget_show (frame);
 
-	gis->_priv->gil = gnome_icon_list_new(ICON_SIZE+30, FALSE);
-	gtk_widget_set_usize(gis->_priv->gil,350,300);
-	gnome_icon_list_set_selection_mode(GNOME_ICON_LIST(gis->_priv->gil),
-					    GTK_SELECTION_SINGLE);
-
-	vadj = gtk_layout_get_vadjustment(GTK_LAYOUT(gis->_priv->gil));
-	sb = gtk_vscrollbar_new(vadj);
+	sb = gtk_vscrollbar_new(NULL);
 	gtk_box_pack_end(GTK_BOX(box),sb,FALSE,FALSE,0);
 	gtk_widget_show(sb);
 
+	gis->_priv->gil = gnome_icon_list_new(ICON_SIZE+30,
+					      gtk_range_get_adjustment (GTK_RANGE (sb)),
+					      FALSE);
+	gtk_widget_set_size_request (gis->_priv->gil, 350, 300);
+	gnome_icon_list_set_selection_mode(GNOME_ICON_LIST(gis->_priv->gil),
+					    GTK_SELECTION_SINGLE);
+
 	gtk_container_add (GTK_CONTAINER (frame), gis->_priv->gil);
 	gtk_widget_show(gis->_priv->gil);
 
+	g_signal_connect (gis->_priv->gil, "destroy",
+			  G_CALLBACK (gtk_widget_destroyed),
+			  &(gis->_priv->gil));
+
 	gis->_priv->file_list = NULL;
 }
 
@@ -135,7 +142,7 @@ gnome_icon_selection_new (void)
 {
   GnomeIconSelection * gis;
   
-  gis = gtk_type_new (GNOME_TYPE_ICON_SELECTION);
+  gis = g_object_new (GNOME_TYPE_ICON_SELECTION, NULL);
 
   return GTK_WIDGET (gis);
 }
@@ -145,37 +152,44 @@ gnome_icon_selection_destroy (GtkObject *object)
 {
 	GnomeIconSelection *gis;
 
-	g_return_if_fail(object != NULL);
-	g_return_if_fail(GNOME_IS_ICON_SELECTION(object));
-
-	/* remember, destroy can be run multiple times! */
+	g_return_if_fail (object != NULL);
+	g_return_if_fail (GNOME_IS_ICON_SELECTION(object));
 	
-	gis = GNOME_ICON_SELECTION(object);
+	gis = GNOME_ICON_SELECTION (object);
 
-	/*clear our data if we have some*/
-	if(gis->_priv->file_list) {
-		g_list_foreach(gis->_priv->file_list,(GFunc)g_free,NULL);
-		g_list_free(gis->_priv->file_list);
-		gis->_priv->file_list = NULL;
-	}
+	if (gis->_priv->load_idle != 0)
+		g_source_remove (gis->_priv->load_idle);
+	gis->_priv->load_idle = 0;
 
-	GNOME_CALL_PARENT_HANDLER (GTK_OBJECT_CLASS, destroy, (object));
+	if (gis->_priv->load_loop != NULL &&
+	    g_main_loop_is_running (gis->_priv->load_loop))
+		g_main_loop_quit (gis->_priv->load_loop);
+
+	GNOME_CALL_PARENT (GTK_OBJECT_CLASS, destroy, (object));
 }
 
+
 static void
 gnome_icon_selection_finalize (GObject *object)
 {
 	GnomeIconSelection *gis;
 
-	g_return_if_fail(object != NULL);
-	g_return_if_fail(GNOME_IS_ICON_SELECTION(object));
+	g_return_if_fail (object != NULL);
+	g_return_if_fail (GNOME_IS_ICON_SELECTION(object));
 	
-	gis = GNOME_ICON_SELECTION(object);
+	gis = GNOME_ICON_SELECTION (object);
 
-	g_free(gis->_priv);
+	/*clear our data if we have some*/
+	if (gis->_priv->file_list != NULL) {
+		g_list_foreach (gis->_priv->file_list, (GFunc)g_free, NULL);
+		g_list_free (gis->_priv->file_list);
+		gis->_priv->file_list = NULL;
+	}
+
+	g_free (gis->_priv);
 	gis->_priv = NULL;
 
-	GNOME_CALL_PARENT_HANDLER (G_OBJECT_CLASS, finalize, (object));
+	GNOME_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
 }
 
 /**
@@ -185,15 +199,14 @@ gnome_icon_selection_finalize (GObject *object)
  * Description: Adds the default pixmap directory into the selection
  * widget. It doesn't show the icons in the selection until you
  * do #gnome_icon_selection_show_icons.
- *
- * Returns:
  **/
 void
 gnome_icon_selection_add_defaults (GnomeIconSelection * gis)
 {
   gchar *pixmap_dir;
 
-  g_return_if_fail(gis != NULL);
+  g_return_if_fail (gis != NULL);
+  g_return_if_fail (GNOME_IS_ICON_SELECTION (gis));
 
   pixmap_dir = gnome_unconditional_datadir_file("pixmaps");
   
@@ -236,7 +249,7 @@ append_an_icon(GnomeIconSelection * gis, const gchar * path)
 		GdkPixbuf *scaled;
 	        scaled = gdk_pixbuf_scale_simple(pixbuf, w, h,
 						 GDK_INTERP_BILINEAR);
-		gdk_pixbuf_unref(pixbuf);
+		g_object_unref (G_OBJECT (pixbuf));
 		pixbuf = scaled;
 
 		/* sanity */
@@ -248,12 +261,12 @@ append_an_icon(GnomeIconSelection * gis, const gchar * path)
 	pos = gnome_icon_list_append_pixbuf(GNOME_ICON_LIST(gis->_priv->gil),
 					    pixbuf, path, base);
 	g_free(base);
-	gdk_pixbuf_unref(pixbuf); /* I'm so glad that gdk-pixbuf has eliminated the former lameness of imlib! :) */
+	g_object_unref (G_OBJECT (pixbuf)); /* I'm so glad that gdk-pixbuf has eliminated the former lameness of imlib! :) */
 }
 
 static int sort_file_list( gconstpointer a, gconstpointer b)
 {
-	return strcmp( (gchar *)a, (gchar *)b );
+	return strcoll( (gchar *)a, (gchar *)b );
 }
 
 /**
@@ -264,8 +277,6 @@ static int sort_file_list( gconstpointer a, gconstpointer b)
  * Description: Adds the icons from the directory @dir to the
  * selection widget. It doesn't show the icons in the selection
  * until you do #gnome_icon_selection_show_icons.
- *
- * Returns:
  **/
 void
 gnome_icon_selection_add_directory (GnomeIconSelection * gis,
@@ -274,8 +285,9 @@ gnome_icon_selection_add_directory (GnomeIconSelection * gis,
   struct dirent * de;
   DIR * dp;
 
-  g_return_if_fail(gis != NULL);
-  g_return_if_fail(dir != NULL);
+  g_return_if_fail (gis != NULL);
+  g_return_if_fail (GNOME_IS_ICON_SELECTION (gis));
+  g_return_if_fail (dir != NULL);
 
   if ( ! g_file_test (dir, G_FILE_TEST_IS_DIR)) {
 	  g_warning(_("GnomeIconSelection: '%s' does not exist or is not "
@@ -292,18 +304,26 @@ gnome_icon_selection_add_directory (GnomeIconSelection * gis,
 
   while ( (de = readdir(dp)) != NULL ) {
     const char *mimetype;
-#ifdef GNOME_ENABLE_DEBUG
-    g_print("File: %s\n", de->d_name);
-#endif
+    char *uri;
+    GnomeVFSFileInfo *info;
+    gchar *full_path;
+    
     if ( *(de->d_name) == '.' ) continue; /* skip dotfiles */
 
-    mimetype = gnome_vfs_mime_type_from_name(de->d_name);
+    full_path = g_build_filename (dir, de->d_name, NULL);
+    uri = g_filename_to_uri (full_path, "localhost", NULL);
+    info = gnome_vfs_file_info_new ();
+    g_free (full_path);
+    /* FIXME: OK to do synchronous I/O here? */
+    gnome_vfs_get_file_info (uri, info, GNOME_VFS_FILE_INFO_GET_MIME_TYPE | 
+			     GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
+    mimetype = info->mime_type;
+    g_free (uri);
+
     if (mimetype != NULL &&
 	strncmp(mimetype, "image", strlen("image")) == 0 ) {
       gchar * full_path = g_build_filename (dir, de->d_name, NULL);
-#ifdef GNOME_ENABLE_DEBUG
-      g_print("Full path: %s\n", full_path);
-#endif
+
       if (g_file_test (full_path, G_FILE_TEST_IS_REGULAR)) {
 	      /* Image filename, exists, regular file, go for it. */
 	      gis->_priv->file_list =
@@ -313,15 +333,46 @@ gnome_icon_selection_add_directory (GnomeIconSelection * gis,
       }
       g_free(full_path);
     }
+    gnome_vfs_file_info_unref (info);
   }
 
   closedir(dp);
 }
 
-static void
-set_flag(GtkWidget *w, int *flag)
+static gboolean
+load_idle_func (gpointer data)
 {
-	*flag = TRUE;
+	GnomeIconSelection * gis = data;
+	GList * list = gis->_priv->file_list;
+
+	if (list == NULL) {
+		if (gis->_priv->load_loop != NULL &&
+		    g_main_loop_is_running (gis->_priv->load_loop))
+			g_main_loop_quit (gis->_priv->load_loop);
+		gis->_priv->load_idle = 0;
+		return FALSE;
+	}
+
+	GDK_THREADS_ENTER();
+
+	append_an_icon (gis, list->data);
+
+	g_free (list->data);
+	gis->_priv->file_list = g_list_remove_link (gis->_priv->file_list, list);
+	g_list_free_1 (list);
+
+	/* If icons were added while adding */
+	if (gis->_priv->load_i > gis->_priv->load_file_count)
+		gis->_priv->load_file_count = gis->_priv->load_i;
+
+	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (gis->_priv->load_progressbar),
+				       (double)gis->_priv->load_i / gis->_priv->load_file_count);
+
+	gis->_priv->load_i ++;
+
+	GDK_THREADS_LEAVE();
+
+	return TRUE;
 }
 
 /**
@@ -333,105 +384,81 @@ set_flag(GtkWidget *w, int *flag)
  * #gnome_icon_selection_add_directory. Before this function
  * is called the icons aren't actually added to the listing 
  * and can't be picked by the user.
- *
- * Returns:
  **/
 void
 gnome_icon_selection_show_icons (GnomeIconSelection * gis)
 {
-  GtkWidget *label;
-  GtkWidget *progressbar;
-  int file_count, i;
-  int local_dest;
-  int was_destroyed = FALSE;
+	GtkWidget *label;
+	GtkWidget *progressbar;
 
-  g_return_if_fail(gis != NULL);
-  if(!gis->_priv->file_list) return;
+	g_return_if_fail (gis != NULL);
+	g_return_if_fail (GNOME_IS_ICON_SELECTION (gis));
 
-  file_count = g_list_length(gis->_priv->file_list);
-  i = 0;
+	/* Nothing to load */
+	if (gis->_priv->file_list == NULL)
+		return;
 
-  /* Locate previous progressbar/label,
-   * if previously called. */
-  progressbar = label = NULL;
-  progressbar = gtk_object_get_user_data(GTK_OBJECT(gis));
-  if (progressbar)
-         label = gtk_object_get_user_data(GTK_OBJECT(progressbar));
+	/* We're already loading! */
+	if (gis->_priv->load_loop != NULL)
+		return;
 
-  if (!label && !progressbar) {
-         label = gtk_label_new(_("Loading Icons..."));
-         gtk_box_pack_start(GTK_BOX(gis->_priv->box),label,FALSE,FALSE,0);
-         gtk_widget_show(label);
+	label = gtk_label_new (_("Loading Icons..."));
+	_add_atk_relation (GTK_WIDGET (gis), label,
+			   ATK_RELATION_LABELLED_BY, ATK_RELATION_LABEL_FOR);
+	gtk_box_pack_start (GTK_BOX (gis->_priv->box),
+			    label, FALSE, FALSE, 0);
+	gtk_widget_show (label);
+	g_signal_connect (label, "destroy",
+			  G_CALLBACK (gtk_widget_destroyed),
+			  &label);
 
-         progressbar = gtk_progress_bar_new();
-         gtk_box_pack_start(GTK_BOX(gis->_priv->box),progressbar,FALSE,FALSE,0);
-         gtk_widget_show(progressbar);
+	progressbar = gtk_progress_bar_new ();
+	_add_atk_relation (progressbar, label,
+			   ATK_RELATION_LABELLED_BY, ATK_RELATION_LABEL_FOR);
+	gtk_box_pack_start (GTK_BOX (gis->_priv->box),
+			    progressbar, FALSE, FALSE, 0);
+	gtk_widget_show (progressbar);
+	g_signal_connect (progressbar, "destroy",
+			  G_CALLBACK (gtk_widget_destroyed),
+			  &progressbar);
 
-         /* attach label to progressbar, progressbar to gis 
-          * for recovery if show_icons() called again */
-         gtk_object_set_user_data(GTK_OBJECT(progressbar), label);
-         gtk_object_set_user_data(GTK_OBJECT(gis), progressbar);
-  } else {
-         if (!label && progressbar) g_assert_not_reached();
-         if (label && !progressbar) g_assert_not_reached();
-  }
-         
-  gnome_icon_list_freeze(GNOME_ICON_LIST(gis->_priv->gil));
+	gnome_icon_list_freeze (GNOME_ICON_LIST (gis->_priv->gil));
 
-  /* this can be set with the stop_loading method to stop the
-     display in the middle */
-  gis->_priv->stop_loading = FALSE;
-  
-  /*bind destroy so that we can bail out of this function if the whole thing
-    was destroyed while doing the main_iteration*/
-  local_dest = gtk_signal_connect(GTK_OBJECT(gis),"destroy",
-				  GTK_SIGNAL_FUNC(set_flag),
-				  &was_destroyed);
+	/* Ref to avoid object disappearing from under us */
+	g_object_ref (G_OBJECT (gis));
 
-  while (gis->_priv->file_list) {
-	  GList * list = gis->_priv->file_list;
-	  append_an_icon(gis, list->data);
-	  g_free(list->data);
-	  gis->_priv->file_list = g_list_remove_link(gis->_priv->file_list,list);
-	  g_list_free_1(list);
+	gis->_priv->load_loop = g_main_loop_new (NULL, FALSE);
+	gis->_priv->load_i = 0;
+	gis->_priv->load_file_count = g_list_length (gis->_priv->file_list);
+	gis->_priv->load_progressbar = progressbar;
 
-	  gtk_progress_bar_update (GTK_PROGRESS_BAR (progressbar),
-				   (float)i / file_count);
-	  /* FIXME: this should be done either
-	   * 1) asynchroniously
-	   * 2) with a separate main loop and not the gtk one */
-	  while ( gtk_events_pending() ) {
-                  gtk_main_iteration();
+	if (gis->_priv->load_idle != 0)
+		g_source_remove (gis->_priv->load_idle);
+	gis->_priv->load_idle = g_idle_add (load_idle_func, gis);
 
-                  /*if the gis was destroyed from underneath us ... bail out*/
-                  if(was_destroyed) 
-                          return;
-                  
-                  if(gis->_priv->stop_loading)
-                          goto out;
-	  }
-	  
-	  i++;
-  }
+	GDK_THREADS_LEAVE ();  
+	g_main_loop_run (gis->_priv->load_loop);
+	GDK_THREADS_ENTER ();  
 
- out:
-  
-  gtk_signal_disconnect(GTK_OBJECT(gis),local_dest);
+	if (gis->_priv->load_idle != 0)
+		g_source_remove (gis->_priv->load_idle);
+	gis->_priv->load_idle = 0;
 
-  gnome_icon_list_thaw(GNOME_ICON_LIST(gis->_priv->gil));
+	if (gis->_priv->load_loop != NULL)
+		g_main_loop_unref (gis->_priv->load_loop);
+	gis->_priv->load_loop = NULL;
 
-  progressbar = label = NULL;
-  progressbar = gtk_object_get_user_data(GTK_OBJECT(gis));
-  if (progressbar)
-         label = gtk_object_get_user_data(GTK_OBJECT(progressbar));
-  if (progressbar)
-	  gtk_widget_destroy(progressbar);
-  if (label)
-	  gtk_widget_destroy(label);
+	gis->_priv->load_progressbar = NULL;
 
-  /* cleanse gis of evil progressbar/label ptrs */
-  /* also let previous calls to show_icons() know that rendering is done. */
-  gtk_object_set_user_data(GTK_OBJECT(gis), NULL);
+	if (gis->_priv->gil != NULL)
+		gnome_icon_list_thaw (GNOME_ICON_LIST (gis->_priv->gil));
+
+	if (progressbar != NULL)
+		gtk_widget_destroy (progressbar);
+	if (label != NULL)
+		gtk_widget_destroy (label);
+
+	g_object_unref (G_OBJECT (gis));
 }
 
 /**
@@ -443,18 +470,18 @@ gnome_icon_selection_show_icons (GnomeIconSelection * gis)
  * harmless, it should be used say if the dialog was hidden
  * or when we want to quickly stop loading the images to do
  * something else without destroying the icon selection object.
- * The ramaining icons can be shown by
+ * The remaining icons can be shown by
  * #gnome_icon_selection_show_icons.
- *
- * Returns:
  **/
 void
-gnome_icon_selection_stop_loading(GnomeIconSelection * gis)
+gnome_icon_selection_stop_loading (GnomeIconSelection * gis)
 {
-	g_return_if_fail(gis != NULL);
-	g_return_if_fail(GNOME_IS_ICON_SELECTION(gis));
+	g_return_if_fail (gis != NULL);
+	g_return_if_fail (GNOME_IS_ICON_SELECTION (gis));
 	
-	gis->_priv->stop_loading = TRUE;
+	if (gis->_priv->load_loop != NULL &&
+	    g_main_loop_is_running (gis->_priv->load_loop))
+		g_main_loop_quit (gis->_priv->load_loop);
 }
 
 /**
@@ -465,8 +492,6 @@ gnome_icon_selection_stop_loading(GnomeIconSelection * gis)
  * Description: Clear the currently shown icons, the ones
  * that weren't shown yet are not cleared unless the not_shown
  * parameter is given, in which case even those are cleared.
- *
- * Returns:
  **/
 void
 gnome_icon_selection_clear (GnomeIconSelection * gis,
@@ -504,7 +529,8 @@ gnome_icon_selection_get_icon (GnomeIconSelection * gis,
 {
 	GList * sel;
 
-	g_return_val_if_fail(gis != NULL, NULL);
+	g_return_val_if_fail (gis != NULL, NULL);
+	g_return_val_if_fail (GNOME_IS_ICON_SELECTION (gis), NULL);
 
 	sel = gnome_icon_list_get_selection(GNOME_ICON_LIST(gis->_priv->gil));
 	if (sel != NULL) {
@@ -527,8 +553,6 @@ gnome_icon_selection_get_icon (GnomeIconSelection * gis,
  *
  * Description: Selects the icon @filename. This icon must have
  * already been added and shown * (see @gnome_icon_selection_show_icons)
- *
- * Returns:
  **/
 void
 gnome_icon_selection_select_icon (GnomeIconSelection * gis,
@@ -538,8 +562,9 @@ gnome_icon_selection_select_icon (GnomeIconSelection * gis,
 	gint pos;
 	gint icons;
 
-	g_return_if_fail(gis != NULL);
-	g_return_if_fail(filename != NULL);
+	g_return_if_fail (gis != NULL);
+	g_return_if_fail (GNOME_IS_ICON_SELECTION (gis));
+	g_return_if_fail (filename != NULL);
 
 	gil = GNOME_ICON_LIST(gis->_priv->gil);
 	icons = gnome_icon_list_get_num_icons(gil);
